@@ -88,12 +88,9 @@ class GeomCustom(StructuredAttribut):
 
     (self.numbVerts, self.numbTVerts, self.numbTSets, self.numbElems, self.numbSets) = self._geom.readInternal(st)
 
-    # Control given lengths vs found
-    if self.numbVerts != len(self._geom.coordList):
-      self.numbVerts = len(self._geom.coordList)
-
-    if self.numbTVerts != len(self._geom.texList):
-      self.numbTVerts = len(self._geom.texList)
+    # Ensure consistency of lengths vs found
+    self.numbVerts = len(self._geom.coordList)
+    self.numbTVerts = len(self._geom.texList)
 
   def applyDelta(self, dlt, factor, mapPt, comCoord:'set of integer'):
     demifactor = factor / 2.0
@@ -325,5 +322,3 @@ class FigureResFile(SimpleAttribut):
   def getPath(self, poserRootDir):
     return getRealPath(poserRootDir, self.getValue())
 
-
-
