@@ -1,14 +1,16 @@
 # -*- coding: utf-8 -*-
 '''
 Comme functions and consts for Poser file analysis.
-Include Poser's vocabulary up to Poser 9 (partially for Poser 11)
+Include Poser's vocabulary up to Poser 9 (partially for Poser 11, more or less complete Poser 12)
 
 Created on 1 mai 2020
 @author: olivier
 '''
-import sys, traceback
+import traceback
+import gzip
 import os.path
 import copy
+import logging
 from collections import namedtuple
 import platform
 import importlib
@@ -18,6 +20,7 @@ from xlrd.sheet import ctype_text, XL_CELL_TEXT, XL_CELL_NUMBER, XL_CELL_EMPTY
 
 from langutil import File, C_FAIL
 
+
 # Test & Set
 #    * @param s
 #    * @param old
@@ -267,9 +270,9 @@ class PoserConst(object):
   POSER_V5  = "5"
   POSER_V6  = "6"
   POSER_V7  = "7"
-  POSER_V9  = "9"
-  POSER_V11, POSER_V11f  = "11", 11.0
-  POSER_V12 = "12"
+  POSER_V9,  POSER_V9f  = "9",   9.0
+  POSER_V11, POSER_V11f = "11", 11.0
+  POSER_V12, POSER_V12f = "12", 12.0
 
   # Poser file extensions.
   # ORDER shall not be changed !
@@ -381,6 +384,15 @@ class PoserToken:
   E_file = PToken("file", False) # v5
   E_fireflyRoot = PToken("fireflyRoot", False, None, False) # v11
   E_superflyRoot = PToken("superflyRoot", False, None, False) # v11
+  
+  E_compoundOutputsPos = PToken("compoundOutputsPos", False, None, False) # v12
+  E_compoundInputsPos = PToken("compoundInputsPos", False, None, False) # v12
+  E_compoundShowPreview = PToken("compoundShowPreview", False, None, False) # v12
+  # E_output = PToken("output", True, None, False) # v12
+  E_selectedOutput = PToken("selectedOutput", False, None, False) # v12
+  E_output = PToken("output", True, 'Output', False) # v12
+  E_exposedAs = PToken("exposedAs") # v12  
+  
   E_backfaceCull = PToken("backfaceCull", False, None, False) # v5
   E_visibleInReflections = PToken("visibleInReflections", False, None, False) # v5
   E_visibleInRender = PToken("visibleInRender", False, None, False) # v5
@@ -658,6 +670,10 @@ class PoserToken:
   E_camAutoCenterX = PToken("camAutoCenterX", True, 'GenericTransform', False)
   E_camAutoCenterY = PToken("camAutoCenterY", True, 'GenericTransform', False)
   E_camAutoCenterZ = PToken("camAutoCenterZ", True, 'GenericTransform', False)
+  # Poser12 new Camera channels
+  E_apertureRatio = PToken("apertureRatio", True, 'GenericTransform', False)
+  E_apertureBlades = PToken("apertureBlades", True, 'GenericTransform', False)
+  E_apertureBladesRotation = PToken("apertureBladesRotation", True, 'GenericTransform', False)
 
   E_valueParm = PToken("valueParm", True, 'GenericTransform', False)
   E_xOffsetA = PToken("xOffsetA", True, 'GenericTransform', False, 'xOffsetA')
@@ -783,6 +799,8 @@ class PoserToken:
   E_auxCamera = PToken("auxCamera", False, None, False)
   E_dollyCamera = PToken("dollyCamera", False, None, False)
   E_useCamera = PToken("useCamera", False, None, False)
+  E_categories = PToken("categories", False, None, False) # Poser12
+  E_displayMeasurements = PToken("displayMeasurements", False, None, False) # Poser12
 
   E_illustrationParms = PToken("illustrationParms", True, 'StructuredAttribut', False)
   E_combineGradient = PToken("combineGradient", False, None, False)
@@ -813,8 +831,76 @@ class PoserToken:
   E_jointX = PToken("jointX", True, 'GenericTransform', False)
   E_jointY = PToken("jointY", True, 'GenericTransform', False)
   E_jointZ = PToken("jointZ", True, 'GenericTransform', False)
-  E_renderDefaults = PToken("renderDefaults", True, 'StructuredAttribut', False)
+  E_renderDefaults = PToken("renderDefaults", True, 'RenderDefaults', False)
   E_curve = PToken("curve", True, 'GenericTransform', False)
+  
+  # Set of new SuperFly options - Poser 12
+  E_previewAASamples = PToken("previewAASamples", False, None, False)
+  E_diffuseSamples = PToken("diffuseSamples", False, None, False)
+  E_previewMipMaps = PToken("previewMipMaps", False, None, False)
+  E_doPreviewMultisample = PToken("doPreviewMultisample", False, None, False)
+  E_adaptiveThreshold = PToken("adaptiveThreshold", False, None, False)
+  E_doPostFilter = PToken("doPostFilter", False, None, False)
+  E_geometricEdgeLinesOn = PToken("geometricEdgeLinesOn", False, None, False)
+  E_geometricEdgeLinesWelded = PToken("geometricEdgeLinesWelded", False, None, False)
+  E_geometricEdgeLinesDrawLast = PToken("geometricEdgeLinesDrawLast", False, None, False)
+  E_geometricEdgeDistance = PToken("geometricEdgeDistance", False, None, False)
+  E_postFilterParm1 = PToken("postFilterParm1", False, None, False)
+  E_usePostFX = PToken("usePostFX", False, None, False)
+  E_usePreviewRoot = PToken("usePreviewRoot", False, None, False)
+  E_drawZOutline = PToken("drawZOutline", False, None, False)
+  E_passFFNormal = PToken("passFFNormal", False, None, False)
+  E_passFFToonID = PToken("passFFToonID", False, None, False)
+  E_passFFZDepth = PToken("passFFZDepth", False, None, False)
+  E_passFFPoint = PToken("passFFPoint", False, None, False)
+  E_passFFTextureCoordinates = PToken("passFFTextureCoordinates", False, None, False)
+  E_passFFCustom1 = PToken("passFFCustom1", False, None, False)
+  E_passFFCustom2 = PToken("passFFCustom2", False, None, False)
+  E_passFFCustom3 = PToken("passFFCustom3", False, None, False)
+  E_hider = PToken("hider", False, None, False)
+  E_superFlyOptions = PToken("superFlyOptions", True, 'StructuredAttribut', False)
+  E_aaSamples = PToken("aaSamples", False, None, False)
+  E_adaptiveSampling = PToken("adaptiveSampling", False, None, False)
+  E_adaptiveThreshold = PToken("adaptiveThreshold", False, None, False)
+  E_advancedSamplingControls = PToken("advancedSamplingControls", False, None, False)
+  E_causticsReflective = PToken("causticsReflective", False, None, False)
+  E_causticsRefractive = PToken("causticsRefractive", False, None, False)
+  E_depthOfField = PToken("depthOfField", False, None, False)
+  E_deviceType = PToken("deviceType", False, None, False)
+  E_filterGlossy = PToken("filterGlossy", False, None, False)
+  E_glossySamples = PToken("glossySamples", False, None, False)
+  E_maxBounce = PToken("maxBounce", False, None, False)
+  E_maxDiffuseBounce = PToken("maxDiffuseBounce", False, None, False)
+  E_maxGlossyBounce = PToken("maxGlossyBounce", False, None, False)
+  E_maxTransmissionBounce = PToken("maxTransmissionBounce", False, None, False)
+  E_maxVolumeBounce = PToken("maxVolumeBounce", False, None, False)
+  E_meshLightSamples = PToken("meshLightSamples", False, None, False)
+  E_minBounce = PToken("minBounce", False, None, False)
+  E_progressiveRefine = PToken("progressiveRefine", False, None, False)
+  E_sampleAllLightsDirect = PToken("sampleAllLightsDirect", False, None, False)
+  E_sampleAllLightsIndirect = PToken("sampleAllLightsIndirect", False, None, False)
+  E_sampleClampDirect = PToken("sampleClampDirect", False, None, False)
+  E_sampleClampIndirect = PToken("sampleClampIndirect", False, None, False)
+  E_spatialSplit = PToken("spatialSplit", False, None, False)
+  E_subsurfaceSamples = PToken("subsurfaceSamples", False, None, False)
+  E_tangentSpace = PToken("tangentSpace", False, None, False)
+  E_tileSize = PToken("tileSize", False, None, False)
+  E_transmissionSamples = PToken("transmissionSamples", False, None, False)
+  E_transparentBackground = PToken("transparentBackground", False, None, False)
+  E_transparentMaxBounce = PToken("transparentMaxBounce", False, None, False)
+  E_transparentMinBounce = PToken("transparentMinBounce", False, None, False)
+  E_transparentShadows = PToken("transparentShadows", False, None, False)
+  E_volumeMaxSteps = PToken("volumeMaxSteps", False, None, False)
+  E_volumeSamples = PToken("volumeSamples", False, None, False)
+  E_volumeStepSize = PToken("volumeStepSize", False, None, False)
+  E_wrapperBackground = PToken("wrapperBackground", False, None, False)
+  E_deviceID = PToken("deviceID", False, None, False)
+
+  # New Structure at PoserFile level for Poser12
+  E_postFX = PToken("postFX", True, 'StructuredAttribut', False)
+  E_useDenoise = PToken("useDenoise", False, None, False)
+  E_exposure = PToken("exposure", False, None, False)
+  E_saturation = PToken("saturation", False, None, False)
 
   E_sphereZoneProp = PToken("sphereZoneProp", True, 'SphereZoneProp', False)
   E_baseProp = PToken("baseProp", True, 'BaseProp', False)
@@ -888,6 +974,8 @@ class PoserToken:
   E_k = PToken("k", False, 'KSA', False)
   E_sl = PToken("sl", False, None, False)
   E_spl = PToken("spl", False, 'OffSA', False)
+  E_catid = PToken("catid", False, None, False) # Poser12 key frame data
+  
   E_valueOpDeltaAdd = PToken("valueOpDeltaAdd", False, 'ValueOpDelta', False)
   E_valueOpPlus = PToken("valueOpPlus", False, 'ValueOpDelta', False)
   E_valueOpMinus = PToken("valueOpMinus", False, 'ValueOpDelta', False)
@@ -1053,6 +1141,28 @@ class PoserToken:
   E_inLenR = PToken("inLenR", False, None, False)
   E_inLenL = PToken("inLenL", False, None, False)
   E_matrices = PToken("matrices", False, 'SphereMathsRaw', False)
+  
+  # Poser12 Measurement Props
+  E_measurementLine = PToken("measurementLine", True, "MeasurementProp", False)
+  E_measurementAngle = PToken("measurementAngle", True, "MeasurementProp", False)
+  E_measurementCircle = PToken("measurementCircle", True, "MeasurementProp", False)
+  E_measurementPolyline = PToken("measurementPolyline", True, "MeasurementProp", False)
+  E_alignment = PToken("alignment", False, None, False)
+  E_extLineScaleParm = PToken("extLineScaleParm", False, None, False)
+  E_calloutStyle = PToken("calloutStyle", False, None, False)
+  E_points = PToken("points", True, "StructuredAttribut", False)
+  E_gluePoint = PToken("gluePoint", True, "GluePoint", False)
+  E_polyInd = PToken("polyInd")
+  E_baryPolyInd = PToken("baryPolyInd")
+  E_baryCoords = PToken("baryCoords")
+  E_splineType = PToken("splineType", False, None, False)
+  E_closed = PToken("closed", False, None, False)
+  E_slack = PToken("slack", False, None, False)
+
+  E_enumValue = PToken("enumValue") # Optional in NodeInput
+
+
+
 
   # Channel initialization:
   #    PoserType, isStatic, min, max, initValue, order XYZ, ZXY, YZX, TrackingScale
@@ -1199,6 +1309,37 @@ def create(ptk, name):
   return sa
 
 
+class PoserWriter:
+  ''' File Writer with Poser File Version '''
+  
+  def __init__(self, fn, compressed=False, version=PoserConst.POSER_V9f):
+    ''' Open a writing file (gzip or not).
+    By default, POSER V9 language will be used.
+    '''
+    self._compressed = False
+    self._fn = fn
+    self._version = version
+    
+    
+    if compressed:
+      logging.info(f"File[{fn}]: writing/zipping")
+      self._prn = gzip.open(fn, 'wt')
+    else:
+      logging.info(f"File[{fn}] : writing")
+      self._prn = open(fn, 'w')
+    
+  def write(self, s):
+    return self._prn.write(s)
+  
+  def close(self):
+    logging.info(f"File[{self._fn}]: {os.path.getsize(self._fn)} bytes")
+    self._prn.close()
+    self._prn = None
+
+  def getVersion(self):
+    return self._version
+
+
 #  private static void readXLSLine(HSSFRow row, String[][] desttab, int rowid, int nbcols)
 def __readXLSLine(sheet, desttab, rowid, nbcols):
   desttab.append([ ])
@@ -1257,20 +1398,6 @@ def readXLSFile(filename):
   return tabres
 
 
-if __name__ == '__main__':
-  for tk in LangList[1:]:
-      attr = create(tk, 'toto')
-      print('Object({0:s}) size={1:d}'.format(tk.token, sys.getsizeof(attr)))
-      try:
-        attr.write(sys.stdout,'')
-      except Exception as e:
-        if e.args[0]==PoserConst.C_EX_NOTCALL:
-          attr.writeDef(sys.stdout, '')
-          attr.writeData(sys.stdout,'')
-        else:
-          traceback.print_last()
-
-
 
 
 
