@@ -5,7 +5,7 @@ import os, gzip
 from langutil import C_OK, C_FAIL, C_ERROR, version2float
 from pypos3d.wftk.PoserFileParser import PoserFileParser, PZ3_FNAT
 
-from pypos3d.pftk.PoserBasic import getRealPath, index, PoserConst, PoserToken , isCompressed, Lang, getPoserFileType, create
+from pypos3d.pftk.PoserBasic import getRealPath, index, PoserConst, PoserToken , isCompressed, Lang, getPoserFileType, create, PoserWriter
 from pypos3d.pftk.StructuredAttribut import GenericTransform, NodeInput, PoserMaterial
 from pypos3d.pftk.GeomCustom import GeomCustom
 from pypos3d.pftk.PoserMeshed import PoserProp, Camera, Light
@@ -80,7 +80,21 @@ class PoserFile(object):
                          PoserToken.E_scale, PoserToken.E_scaleX, PoserToken.E_scaleY, PoserToken.E_scaleZ,
                          PoserToken.E_visibility
                          }
-
+  # Set of Poser Types of All Props
+  PROPSTYPES = { PoserToken.E_controlProp, 
+                 PoserToken.E_baseProp, 
+                 PoserToken.E_hairProp, 
+                 PoserToken.E_magnetDeformerProp,
+                 PoserToken.E_sphereZoneProp,
+                 PoserToken.E_curveProp,
+                 PoserToken.E_waveDeformerProp,
+                 PoserToken.E_coneForceFieldProp,
+                 PoserToken.E_controlHandleProp,
+                 PoserToken.E_measurementLine,
+                 PoserToken.E_measurementAngle,
+                 PoserToken.E_measurementCircle, 
+                 PoserToken.E_measurementPolyline 
+                 }     
 
   def __init__(self, fn, createLinks=False):
     ''' Create a PoserFile from a file.
@@ -108,7 +122,7 @@ class PoserFile(object):
 
     #  Poser File Version
     self._version = None
-    self._floatVersion = 9.0 # Default version is set to Poser 9
+    self._floatVersion = PoserConst.POSER_V9f # Default version is set to Poser 9
 
     #  Poser File PZ3 movieinfo
     self._movieinfo = None
@@ -142,6 +156,7 @@ class PoserFile(object):
     self._illustrationParms = None
     self._renderDefaults = None
     self._faceRoom = None
+    self._postFX = None # new Poser12 structure to describe post Rendering FX
 
     # Channels link Dictionnary
     self._gtlnk = {}
@@ -194,15 +209,18 @@ class PoserFile(object):
     self._name = s
 
   #
-  # Return the Poser Language Version as an PoserObject
-  #
-  def getVersion(self): return self._version
+  def getVersion(self):
+    ''' Return the Poser Language Version as an PoserObject.
+    May be None, if no file version is given
+    ''' 
+    return self._version
 
-#  def isVersion(self, cmpVers):
-#    if not self._versionStr:
-#      sa = self._version.findAttribut(PoserToken.E_number)
-#      self._versionStr = sa.getValue() if sa else ''
-#    return False if self._versionStr==None else (self._versionStr==cmpVers)
+  #
+  def getFloatVersion(self):
+    ''' Return the Poser Language Version as float for comparison purpose
+    Default is POSER_V9f
+    ''' 
+    return self._floatVersion
 
   def isBetterVersion(self, v):
     ''' Test if current PoserFile version is better (younger) then given one. '''
@@ -383,15 +401,7 @@ class PoserFile(object):
 
     # elif pot==PoserToken.E_figure: Managed in read() as of V0.2
     # Manage other Props : hairProp, controlProp, Magnets
-    elif pot==PoserToken.E_controlProp or \
-         pot==PoserToken.E_baseProp or \
-         pot==PoserToken.E_hairProp or \
-         pot==PoserToken.E_magnetDeformerProp or \
-         pot==PoserToken.E_sphereZoneProp or \
-         pot==PoserToken.E_curveProp or \
-         pot==PoserToken.E_waveDeformerProp or \
-         pot==PoserToken.E_coneForceFieldProp or \
-         pot==PoserToken.E_controlHandleProp :
+    elif pot in PoserFile.PROPSTYPES:      
       pp = po
       pp.setPoserFile(self)
       if pp.getName() == "FocusDistanceControl":
@@ -425,6 +435,9 @@ class PoserFile(object):
     elif pot==PoserToken.E_doc:
       self._doc = po
       return self._doc
+    elif pot==PoserToken.E_postFX:
+      self._postFX = po
+      return self._postFX
     elif pot==PoserToken.E_setGeomHandlerOffset:
       return None
     #elif pot==PoserToken.E_readScript: #  TODO : Read Other Definition Script
@@ -498,11 +511,8 @@ class PoserFile(object):
         if actorPropName == a.getName():
           return a
 
-    for pp in self._lstProp + self._lstLight + self._lstCamera:
-      if actorPropName == pp.getName():
-        return pp
-
-    return None
+    return next( ( pp for pp in ([ self._ground, ] if self._ground else []) + self._lstProp + self._lstLight + self._lstCamera \
+                   if actorPropName == pp.getName()), None)
 
   #
   # Return the list of PoserMeshedObject (actor, prop, hairprop)
@@ -519,7 +529,7 @@ class PoserFile(object):
         [ a for fig in self._lstFigure for a in fig.getActors() ]
 
   #
-  def writeFile(self, fn):
+  def writeFile(self, fn, version=PoserConst.POSER_V9f):
     ''' Write PoserFile in text format (without compression)
     # RENAMED Java write --> Python writeFile
 
@@ -527,7 +537,10 @@ class PoserFile(object):
     ----------
     fn : str
       Full path name
-
+      
+    version : float, default PoserConst.POSER_V9f
+      Float value of the Poser Syntax
+      
     Returns
     -------
     int
@@ -535,17 +548,16 @@ class PoserFile(object):
     '''
     ret = C_OK
     try:
-      logging.info(f"File[{fn}] : writing")
-      prn = open(fn, 'w')
+      # Use the new Class PoserWriter to manage Syntax version
+      prn = PoserWriter(fn, False, version)
       self.write(prn)
       prn.close()
-      logging.info(f"File[{fn}]: {os.path.getsize(fn)} bytes")
     except OSError:
       ret = C_ERROR
     return ret
 
   #
-  def writeZ(self, fn):
+  def writeZ(self, fn, version=PoserConst.POSER_V9f):
     ''' Write PoserFile in Zlib format (text stream compressed)
 
     Parameters
@@ -560,34 +572,36 @@ class PoserFile(object):
     '''
     ret = C_OK
     try:
-      logging.info(f"File[{fn}]: writing/zipping")
-      out = gzip.open(fn, 'wt')
+      # Use the new Class PoserWriter to manage Syntax version
+      out = PoserWriter(fn, True, version)
       self.write(out)
       out.close()
-      logging.info(f"File[{fn}]: {os.path.getsize(fn)} bytes")
     except OSError:
       ret = C_ERROR
     return ret
 
 
-  def save(self, fn):
+  def save(self, fn, version=PoserConst.POSER_V9f):
     ''' Write a PoserFile in text mode or in compressed mode.
     Use the extension to find the right mode.
 
     Parameters
     ----------
-    fn : str
+    fn: str
       Full path name
 
+    version: float
+      Poser Language version to use.
+    
     Returns
     -------
     int
       C_OK write without error, C_ERROR a write error has occurred
     '''
     if isCompressed(fn):
-      ret = self.writeZ(fn)
+      ret = self.writeZ(fn, version=version)
     else:
-      ret = self.writeFile(fn)
+      ret = self.writeFile(fn, version=version)
 
     return ret
 
@@ -614,7 +628,13 @@ class PoserFile(object):
     #  Write Definition of Lights
     #  Write Definition of Cameras
     #  Write Definition of Figures
-    allParts = [ x for x in [ self._ground, self._universe, self._focusDistanceControl ] if isPZ3  and x ]
+    # For Poser12, write UNIVERSE first
+    if fw.getVersion()>=PoserConst.POSER_V12f:
+      # Just to ease comparison, start with UNIVERSE actor in Poser 12
+      allParts = [ x for x in [ self._universe, self._focusDistanceControl, self._ground ] if isPZ3  and x ]
+    else:
+      allParts = [ x for x in [ self._ground, self._universe, self._focusDistanceControl ] if isPZ3  and x ]
+      
     allParts += self._lstProp + self._lstLight + self._lstCamera + self._lstFigure
     for part in allParts:
       part.writeDef(fw, "")
@@ -634,7 +654,7 @@ class PoserFile(object):
       self._doc.write(fw, "")
 
     if isPZ3:
-      for attr in [ self._illustrationParms, self._renderDefaults, self._faceRoom ]:
+      for attr in [ self._illustrationParms, self._renderDefaults, self._faceRoom, (self._postFX if fw.getVersion()>=PoserConst.POSER_V12f else None) ]:
         if attr:
           attr.write(fw, "")
 
