@@ -1083,7 +1083,6 @@ class PoserMeshedObject(StructuredAttribut):
     elif tokenID==PoserToken.E_subdivideWithFigure:
       self._subdivideWithFigure = (val[0] == "1")
 
-
     else:
       logging.warning("Unexpected Direct: (%s) %s in (%s)", tokenID.token, val, self.getName())
 
@@ -1168,7 +1167,10 @@ class PoserMeshedObject(StructuredAttribut):
             self.addAttribut(sa)
         except KeyError:  #  Mot inconnu
           logging.info("Line[%s] - Unknown word:%s", st.lineno(), cn)
-
+          
+        # except AttributeError as ae:
+        #   logging.info("Line[%s] - Attributerror:%s", st.lineno(), cn)
+          
       elif code==PoserFileParser.TT_RIGHTBRACKET:
         if inchannels:
           inchannels = False
@@ -1477,11 +1479,6 @@ class ControlHandleProp(PoserActor):
     else:
       super(ControlHandleProp, self).setDirect(tokenID, val)
 
-  def write(self, fw, pfx):
-    if self._actorToControl:
-      fw.write(f"{pfx}{self.getPoserType().token} {self.getValue()}\n")
-    super(ControlHandleProp, self).write(fw, pfx)
-
 #
 # This class represents the hairProp attribut of a Poser File.
 #
@@ -1499,6 +1496,13 @@ class HairProp(PoserProp):
     super(HairProp, self).addAttribut(po)
     self._isSimple = False
 
-
+#
+# Area of Poser12 classes
+#
+class MeasurementProp(PoserProp):
+  ''' Class added for Poser12 '''
+  
+  def __init__(self):
+    super(MeasurementProp, self).__init__()
 
 
