@@ -251,13 +251,16 @@ class NodeInput(StructuredAttribut):
 
     if src:
       self._nodeName = src._nodeName
-      self._name = nodeName if nodeName else src._name
+      self._name     = nodeName if nodeName else src._name
       self._value    = src._value
       self._parmR    = src._parmR
       self._parmG    = src._parmG
       self._parmB    = src._parmB
       self._node     = src._node
       self._file     = src._file
+      self._enumValue = src._enumValue
+      self._exposedAs = src._exposedAs
+
     else:
       self._nodeName = nodeName
       if nodeName:
@@ -269,6 +272,8 @@ class NodeInput(StructuredAttribut):
       self._parmB    = "" #  NO_PARM
       self._node     = "" #  NO_NODE
       self._file     = "" #  "NO_MAP"
+      self._enumValue = None # Optional Poser12 attribut
+      self._exposedAs = None # Optional Poser12 attribut
 
   # Set the name.
   def setName(self, s):
@@ -297,8 +302,12 @@ class NodeInput(StructuredAttribut):
         elif tokenID==PoserToken.E_file:
           val = val.strip()
           self._file = None if val==PoserConst.C_NO_MAPG or val==PoserConst.C_NO_MAP else val
+        elif tokenID==PoserToken.E_enumValue:
+          self._enumValue = val
+        elif tokenID == PoserToken.E_exposedAs:
+          self._exposedAs = RemoveQuotes(val)
         else:
-          logging.warning("Unexpected Direct in NodeInput:%s (%s)",tokenID.token, val)
+          logging.warning("Unexpected word in NodeInput:%s (%s)",tokenID.token if tokenID else cn, val)
 
       elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF):
         break
@@ -315,11 +324,17 @@ class NodeInput(StructuredAttribut):
     fw.write(f"{nextPfx}{{\n")
     fw.write(f"{nextPfx}name \"{self._nodeName}\"\n")
     fw.write(f"{nextPfx}value {self._value}\n")
+    
+    if (self._enumValue is not None) and (fw.getVersion()>= PoserConst.POSER_V12f):
+      fw.write(f"{nextPfx}enumValue {self._enumValue}\n")
+      
     fw.write(f"{nextPfx}parmR {self._parmR if self._parmR else PoserConst.C_NO_PARM}\n")
     fw.write(f"{nextPfx}parmG {self._parmG if self._parmG else PoserConst.C_NO_PARM}\n")
     fw.write(f"{nextPfx}parmB {self._parmB if self._parmB else PoserConst.C_NO_PARM}\n")
     fw.write(f"{nextPfx}node \"{self._node if self._node else PoserConst.C_NO_NODE}\"\n")
     fw.write(f"{nextPfx}file {self._file if self._file else PoserConst.C_NO_MAPG}\n")
+    if (self._exposedAs is not None) and (fw.getVersion()>= PoserConst.POSER_V12f):
+      fw.write(f"{nextPfx}exposedAs \"{self._exposedAs}\"\n")
     fw.write(f"{nextPfx}}}\n")
 
 
@@ -370,7 +385,15 @@ class Node(StructuredAttribut):
       lsrc = src.getLstAttr()
       if lsrc:
         for ni in lsrc:
-          nci = NodeInput(src=ni)
+          if isinstance(ni, NodeInput):
+            nci = NodeInput(src=ni)
+          elif isinstance(ni, Output):
+            nci = Output(src=ni)
+          elif isinstance(ni, ShaderTree):
+            nci = ShaderTree(src=ni)
+          else:
+            nci = SimpleAttribut(ni.getPoserType(), ni.getValue())
+            
           self.addAttribut(nci)
 
     else:
@@ -443,13 +466,25 @@ class Node(StructuredAttribut):
   def getNodeType(self):
     return PoserConst.kNodeTypeCodeBLENDER if self._typeName=='blender' else PoserConst.kNodeTypeCodeIMAGEMAP if self._typeName=='image_map' else C_ERROR
 
-#
-# Shader Tree
 #
 class ShaderTree(StructuredAttribut):
-  def __init__(self):
+  ''' This class represents a ShaderTree. 
+  It's a flat representation of a tree of Nodes
+  '''
+  def __init__(self, src=None):
     super(ShaderTree, self).__init__()
     self.setPoserType(PoserToken.E_shaderTree)
+    
+    if src: # Copy Constructor
+      for n in src.getLstAttr():
+        # It is supposed that ShaderTree only contains Nodes or SimpleAttributes
+        if isinstance(n, Node):
+          nn = Node(src=n)
+        else:
+          nn = SimpleAttribut(n.getPoserType(), n.getValue())
+
+        self.addAttribut(nn)
+
 
   def getNodeByInternalName(self, intName):
     return next( (n for n in self.getLstAttr() if isinstance(n, Node) and ((n._nodeName==intName) or (n._name==intName))), None)
@@ -538,6 +573,58 @@ class ShaderTree(StructuredAttribut):
     ni._node = nodeRef.getName()
 
 
+
+#
+class Output(StructuredAttribut):
+  ''' This class represents the output of a compound node.
+  Introduced in Poser12
+  '''
+  def __init__(self, nodeName='', src=None):
+    super(Output, self).__init__()
+    self._poserType = PoserToken.E_output
+
+    # Copy constructor
+    if src:
+      self._nodeName  = src._nodeName
+      self._name      = nodeName if nodeName else src._name
+      self._exposedAs = src._exposedAs
+    else:
+      self._exposedAs = None
+      self._nodeName  = nodeName
+      if nodeName:
+        self._name = nodeName
+        
+  # Set the name.
+  def setName(self, s):
+    self._name = RemoveQuotes(s)
+    self._nodeName = self._name
+
+  def setDirect(self, tokenID, val):
+    if tokenID == PoserToken.E_name:
+      self._nodeName = RemoveQuotes(val)
+    elif tokenID == PoserToken.E_exposedAs:
+      self._exposedAs = RemoveQuotes(val)
+    else:
+      logging.warning("Unexpected Direct in Output:%s (%s)",tokenID.token, val)
+
+  def write(self, fw, pfx):
+    fw.write(f"{pfx}{self.getPoserType().token} \"{self.getName()}\"\n")
+    nextPfx = pfx + "\t"
+    fw.write(f"{nextPfx}{{\n")
+    if self._name:
+      fw.write(f"{nextPfx}name \"{self._nodeName}\"\n")
+      
+    if self._exposedAs:
+      fw.write(f"{nextPfx}exposedAs \"{self._exposedAs}\"\n")
+
+    fw.write(f"{nextPfx}}}\n")
+
+
+
+
+
+
+
 #
 # material class.
 #
@@ -602,13 +689,7 @@ class PoserMaterial(StructuredAttribut):
         self._ReflectionColor = WFColor(src._ReflectionColor.r, src._ReflectionColor.g, src._ReflectionColor.b, src._ReflectionColor.t)
         self._reflectionStrength = src._reflectionStrength
         if src._shaderTree:
-          self._shaderTree = ShaderTree()
-
-          for n in src._shaderTree.getLstAttr():
-            # It is supposed that ShaderTree only contains Nodes
-            nn = Node(src=n)
-            self._shaderTree.addAttribut(nn)
-
+          self._shaderTree = ShaderTree(src=src._shaderTree)
 
       else: # Illegal source
         raise TypeError()
@@ -776,26 +857,28 @@ class InkyChain(StructuredAttribut):
       fw.write('\n')
     fw.write(f"{nextPfx}}}\n")
 
-#
-#  * This class represents the "keys" entry of every channel definition.
-#  *
-#  * keys
-#  *  {
-#  *  static <0|1>
-#  *  k <long> <real>
-#  *  [sl <0|1>
-#  *  spl|lin|con
-#  *  sm|br]
-#  *  }
-#  *
-#  * static determines whether this channel is animating (=0) or static (=1).
-#  * k denotes the key information. The first values is the frame, starting at 0. The second value is the channel value at this frame.
-#  * sl determines whether or not to loop animation.
-#  * spl|lin|con denote the section interpolation type: spline, linear, constant.
-#  * sm|br represent either a smooth transition or break in the transition.
 #
 class Keys(StructuredAttribut):
-
+  '''  This class represents the "keys" entry of every channel definition.
+   
+   keys
+    {
+    static <0|1>
+    k <long> <real>
+    [sl <0|1>
+    spl|lin|con
+    sm|br
+    [catid int]]
+    }
+  
+   static determines whether this channel is animating (=0) or static (=1).
+   k denotes the key information. The first values is the frame, starting at 0. The second value is the channel value at this frame.
+   sl determines whether or not to loop animation.
+   spl|lin|con denote the section interpolation type: spline, linear, constant.
+   sm|br represent either a smooth transition or break in the transition.
+   catid int: Poser12 attribut 
+  '''  
+  
   def __init__(self, src=None):
     super(Keys, self).__init__()
     self.setPoserType(PoserToken.E_keys)
@@ -833,6 +916,8 @@ class Keys(StructuredAttribut):
 
         elif vc==PoserToken.E_static:
           self._static = (rw[0]=='1')
+        elif vc==PoserToken.E_catid:
+          curK.setCatID(int(rw))
         else:
           logging.warning("L[%d] - Not Accepted :%s", st.lineno(), cn)
 
@@ -886,7 +971,7 @@ class Keys(StructuredAttribut):
   #
   #  Return the list of declared "k" frames.
   #
-  # def getLstKey(self): return self._lstKeys
+  def getKey(self, keyno=0): return self._hshKeys[keyno]
 
   #
   # Convert a GenericTransform to a static one (i.e. no keys)
@@ -944,6 +1029,22 @@ class DocDescription(StructuredAttribut):
 
     fw.write(f"{nextPfx}}}\n")
 
+
+class RenderDefaults(StructuredAttribut):
+  ''' New class to manage RenderDefaults attributes '''
+  def __init__(self):
+    super(RenderDefaults, self).__init__()
+
+  #  Just to Manage 1 direct Attribute
+  def setDirect(self, tokenID, val):
+    if tokenID==PoserToken.E_bumpStrength:
+      self.addAttribut(SimpleAttribut(PoserToken.E_bumpStrength, val))
+    else:
+      logging.warning("Unexpected Direct in RenderDefaults:%s (%s)", tokenID.token, val)
+
+
+
+
 # Description part of a figure.
 #  class FigureDescription(StructuredAttribut): --> inlined in Figure
 
@@ -1966,9 +2067,6 @@ class BoneMap(StructuredAttribut):
   def set(self, key, val):
     self._data.append( (key, val ) )
 
-  #def get(self, key):
- #   return self._data.get(key, None)
-
   # Overloaded reader
   def read(self, st):
     while True:
@@ -2055,6 +2153,65 @@ class SkinWeights(StructuredAttribut):
 
     fw.write(f"{pfx}\t}}\n")
 
+
+class GluePoint(StructuredAttribut):
+  ''' Represents a definition point of a measurement Prop 
+
+      actor GROUND
+      polyInd 1556
+      baryPolyInd 0 2 3
+      baryCoords 0.783707 0.027506 0.188788
+  '''
+  def __init__(self, Name=None):
+    super(GluePoint, self).__init__(Name)
+    
+    # Store Data as Strings for this early implementation
+    self._actor = '' 
+    self._polyInd = ''
+    self._baryPolyInd = ''
+    self._baryCoords = ''
+
+  # Overloaded reader
+  def read(self, st):
+
+    # Read the open brcket
+    code,cn,rw = st.getLine()
+
+    while True:
+      code,cn,rw = st.getLine()
+      if code==PoserFileParser.TT_WORD:
+        vc = Lang.get(cn, None)
+        if vc==PoserToken.E_actor:
+          self._actor = rw
+        elif vc==PoserToken.E_polyInd:
+          self._polyInd = rw
+        elif vc==PoserToken.E_baryPolyInd:
+          self._baryPolyInd = rw
+        elif vc==PoserToken.E_baryCoords:
+          self._baryCoords = rw
+        else:
+          logging.warning("Line[%s] - Not Accepted :%s", st.lineno(), cn)
+          raise ParsingErrorException()
+
+      elif (code==PoserFileParser.TT_RIGHTBRACKET) or (code==PoserFileParser.TT_EOF):
+        break
+      else:
+        raise ParsingErrorException(f'L[{st.lineno()}]: Not Accepted :{cn}')
+    # End of read
+
+  def write(self, fw, pfx):
+    fw.write(f"{pfx}{self.getPoserType().token} {self.getName()}\n")
+    nextPfx = pfx + "\t"
+    fw.write(f"{nextPfx}{{\n")
+    fw.write(f"{nextPfx}actor {self._actor}\n")
+    fw.write(f"{nextPfx}polyInd {self._polyInd}\n")
+    fw.write(f"{nextPfx}baryPolyInd {self._baryPolyInd}\n")
+    fw.write(f"{nextPfx}baryCoords {self._baryCoords}\n")
+    fw.write(f"{pfx}\t}}\n")
+
+
+
+
 #
 # Calculate point mapping table :<br>
 # For each vertex of srcWG, find the closest vertex in the list of reference WaveGeom
