@@ -282,7 +282,11 @@ class GeomGroup(object):
     ''' Add a face defined by indexes of Vertex, Texture Coords and normals if any.
   
     Return the origin coordIdx for convenience
-    '''   
+    '''
+    if len(coordIdx)<3:
+      logging.warning(f'Face {len(self.stripCount)} rejected')
+      return coordIdx
+    
     if not isinstance(coordIdx, array.array):
       coordIdx = array.array('l', coordIdx)
 
